package org.codecop.natural.natstyle;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.codecop.natural.natstyle.pmd.NaturalRule;
import org.codecop.natural.natstyle.pmd.NaturalRuleViolation;
import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;

import net.sourceforge.pmd.Report;
import net.sourceforge.pmd.Report.ProcessingError;
import net.sourceforge.pmd.Rule;
import net.sourceforge.pmd.RuleContext;
import net.sourceforge.pmd.RulePriority;
import net.sourceforge.pmd.RuleViolation;

class NatStyleResultParser extends DefaultHandler {

    private final Map<String, Report> reportsByFile;
    private String currentFile;

    public NatStyleResultParser(Map<String, Report> reportsByFile) {
        this.reportsByFile = reportsByFile;
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) {
        switch (localName) {
        case "NATstyle":
            rootElement(attributes);
            break;
        case "file":
            fileElement(attributes);
            break;
        case "error":
            errorElement(attributes);
            break;
        default:
            System.err.println("Unknown element " + localName);
            break;
        }
    }

    private void rootElement(Attributes attributes) {
        // new report, do nothing
    }

    private void fileElement(Attributes attributes) {
        setCurrentFile(attributes);
        createReportForCurrentFile();
    }

    private void setCurrentFile(Attributes attributes) {
        currentFile = attributes.getValue("location");
    }

    private void createReportForCurrentFile() {
        if (!reportsByFile.containsKey(currentFile)) {
            RuleContext ctx = new RuleContext();
            reportsByFile.put(currentFile, Report.createReport(ctx, currentFile));
        }
    }

    private void errorElement(Attributes attributes) {

        String severity = attributes.getValue("severity");
        Optional<String> line = Optional.ofNullable(attributes.getValue("line"));
        String message = attributes.getValue("message");
        String ruleClass = attributes.getValue("class");
        String name = attributes.getValue("name");

        String parserErrorMarker = "NO NATsyle results, parser ended with error.";
        if (message.startsWith(parserErrorMarker)) {

            String errorsInMessage = message.substring(parserErrorMarker.length()).trim();
            List<String> errors = splitErrors(errorsInMessage);

            for (String error : errors) {
                addError(error);
                addViolation("parser.error", name, severity, extractLineFromError(error), error);
            }
            
        } else {
            addViolation(ruleClass, name, severity, line, message);
        }
    }

    private Optional<String> extractLineFromError(String error) {
        return Optional.of(error.replaceAll(".*\\s+Line:\\s+(\\d+)\\d*$", "$1")).filter(l -> !l.isEmpty());
    }

    private void addError(String error) {
        currentReport().addError(new ProcessingError(error, currentFile));
    }

    private Report currentReport() {
        return reportsByFile.get(currentFile);
    }

    private List<String> splitErrors(String message) {
        List<String> errors = new ArrayList<>();

        Matcher matcher = Pattern.compile("\\d+\\sE_\\d+:").matcher(message);
        int lastFound = 0;
        while (matcher.find()) {
            int end = matcher.start();
            if (lastFound == end) {
                continue;
            }
            errors.add(message.substring(lastFound, end).trim());
            lastFound = end;
        }
        errors.add(message.substring(lastFound, message.length()).trim());

        return errors;
    }

    private void addViolation(String ruleClass, String name, String severity, Optional<String> line, String message) {
        Rule rule = createRuleFor(ruleClass, name, severity);
        RuleViolation violation = createViolationFor(rule, line, message);
        currentReport().addRuleViolation(violation);
    }

    private Rule createRuleFor(String ruleClass, String name, String severity) {
        String rulesetName = creatRulesetNameFrom(ruleClass);
        RulePriority priority = asPriority(severity);
        String url = externalInfoUrlFor(rulesetName);

        return new NaturalRule(name, rulesetName, priority, url);
    }

    private String creatRulesetNameFrom(String ruleClass) {
        return ruleClass.replaceAll("^.*\\.check\\.(.*)\\..*$", "$1");
    }

    private RulePriority asPriority(String severity) {
        Map<String, RulePriority> lookup = new HashMap<>();
        lookup.put("error", RulePriority.HIGH);
        lookup.put("warning", RulePriority.MEDIUM);
        lookup.put("info", RulePriority.LOW);

        if (!lookup.containsKey(severity)) {
            System.err.println("Unknown severity: " + severity);
        }

        return lookup.get(severity);
    }

    private String externalInfoUrlFor(String rulesetName) {
        final String url;
        switch (rulesetName) {
        case "src.source":
            url = "http://techcommunity.softwareag.com/ecosystem/documentation/naturalONE/natONE838/core/using/use-natstyle.htm#use-natstyle-prefover-source";
            break;
        case "src.header":
            url = "http://techcommunity.softwareag.com/ecosystem/documentation/naturalONE/natONE838/core/using/use-natstyle.htm#use-natstyle-prefover-header";
            break;
        case "src.parser":
            url = "http://techcommunity.softwareag.com/ecosystem/documentation/naturalONE/natONE838/core/using/use-natstyle.htm#use-natstyle-prefover-parser";
            break;
        case "err":
            url = "http://techcommunity.softwareag.com/ecosystem/documentation/naturalONE/natONE838/core/using/use-natstyle.htm#use-natstyle-prefover-error";
            break;
        case "res":
            url = "http://techcommunity.softwareag.com/ecosystem/documentation/naturalONE/natONE838/core/using/use-natstyle.htm#use-natstyle-prefover-resource";
            break;
        case "lib":
            url = "http://techcommunity.softwareag.com/ecosystem/documentation/naturalONE/natONE838/core/using/use-natstyle.htm#use-natstyle-prefover-library";
            break;
        default:
            System.err.println("Unknown ruleset name: " + rulesetName);
            url = "http://techcommunity.softwareag.com/ecosystem/documentation/naturalONE/natONE838/core/using/use-natstyle.htm#use-natstyle-prefover";
            break;
        }
        return url;
    }

    private RuleViolation createViolationFor(Rule rule, Optional<String> lineString, String message) {
        int line = parseLine(lineString);

        return new NaturalRuleViolation(rule, message, currentFile, line);
    }

    private int parseLine(Optional<String> lineString) {
        return lineString.map(Integer::parseInt).orElse(0);
    }

}